using ArthaChitra.Model.Common;
using ArthaChitra.Model.IO;
using ArthaChitra.Model.SharpScript;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Serialization;

namespace ArthaChitra.Model.Chart
{

    public class FootprintBar : BarBase
    {
        
        private readonly bool isCumulative;
        public FootprintBar(FootprintBarType footprint, bool isCumulative = false)
        {
            this.isCumulative = isCumulative;
            this.Footprint = footprint;
            this.Items = new ConcurrentDictionary<double, FootprintItem>();
        }
        
        
        internal void PopulatePrice(double high, double low, double bidPrice, double askPrice, double volume)
        {
        
            this.Max = Math.Max(high, this.Max);
            this.Min = Math.Min(low, this.Min);
            
            int steps = (int)((Model.Data.Instrument.RoundToTickSize(high, this.Footprint.TickSize) - Model.Data.Instrument.RoundToTickSize(low, this.Footprint.TickSize)) / this.Footprint.TickSize);
        
            if (steps == 0)
            {
                var price1 = Model.Data.Instrument.RoundToTickSize(high, this.Footprint.TickSize);
                AddPrice(price1, volume, bidPrice, askPrice);
                
                return;
            }
            
            var vol = volume / (double)steps;
            var price = high;
            while (price >= low)
            {
                double roundedPrice = Model.Data.Instrument.RoundToTickSize(price, this.Footprint.TickSize);
                
                AddPrice(roundedPrice, vol, bidPrice, askPrice);
                
                price = Model.Data.Instrument.RoundToTickSize(price - this.Footprint.TickSize, this.Footprint.TickSize);
            }
            
        }
        
        private void AddPrice(double price, double volume, double bidPrice, double askPrice)
        {
            var poc = this.poc;
            
            double bidVol = 0.0d;
            double askVol = 0.0d;
            
            if (bidPrice == 0.0d && askPrice == 0.0d)
            {
                bidVol = askVol = volume / 2.0d;
            }
            else if (bidPrice != 0.0d && bidPrice != askPrice)
            {
                double median = (bidPrice + askPrice) / 2.0d;
                if (price > median)
                {
                    askVol = volume;
                }
                else if (price < median)
                {
                    bidVol = volume;
                }
            }
            
            if (this.Items.TryGetValue(price, out var item))
            {
                item.TotalVolume += volume;
                item.TotalBidVolume += bidVol;
                item.TotalAskVolume += askVol;
                
                if (poc != null && poc != item && item.TotalVolume > poc.TotalVolume)
                {
                    poc.IsPointOfControl = false;
                    this.poc = item;
                    item.IsPointOfControl = true;
                }
                
                if (!this.isCumulative)
                {
                    this.Footprint.POCVolume = Math.Max(item.TotalVolume, this.Footprint.POCVolume);
                }
                return;
            }
            
            var footprintItem = new FootprintItem(price)
            {
                TotalVolume = volume,
                TotalBidVolume = bidVol,
                TotalAskVolume = askVol
            };
            
            
            if (poc == null)
            {
                this.poc = footprintItem;
                footprintItem.IsPointOfControl = true;
            }
            else if (poc.TotalVolume < volume)
            {
                poc.IsPointOfControl = false;
                this.poc = footprintItem;
                footprintItem.IsPointOfControl = true;
            }
            
            if (!this.isCumulative)
            {
                this.Footprint.POCVolume = Math.Max(volume, this.Footprint.POCVolume);
            }
            
            this.Items.TryAdd(price, footprintItem);
        }
        
        public FootprintBarType Footprint { get; }
        
        FootprintItem poc = null;   //current point of control
        public FootprintItem POC => this.poc;
        
        public double Max { get; set; } = double.MinValue;
        public double Min { get; set; } = double.MaxValue;
        
        public ConcurrentDictionary<double, FootprintItem> Items { get; }
    }

    public class FootprintItem
    {
    
        private readonly double price;

        public FootprintItem(double price)
        {
            this.price = price;
        }
   
        public double Price => this.price;
        
        public double TotalVolume { get; set; }
        public double TotalBidVolume { get; set; }
        public double TotalAskVolume { get; set; }
        public bool IsPointOfControl { get; set; }
        
    }

    [DisplayName("Footprint Bars")]
    public class FootprintBarType : BarType
    {
        
        public FootprintBarType() : this(30)
        { }
        
        public FootprintBarType(int barValue)
        {
            this.BarValue = barValue;
            this.profileBar = new FootprintBar(this, isCumulative: true);
        }
        
        protected override void OnBarUpdate(BarUpdateEventArgs args, DateTime time, double open, double high, double low, double close, double bid, double ask, double volume, double oi)
        {
            if (args.Bars.Count == 0)
            {
                this.TickSize = Math.Max(args.Bars.Instrument.TickSize, this.TickSize);
                
                var footprintBar = new FootprintBar(this);
                footprintBar.PopulatePrice(high, low, bid, ask, volume);
                
                AddBar(footprintBar, args, CalculateNewBarTime(time, args), open, high, low, close, volume, oi);
            }
            else
            {
                
                if (time > args.Bars.GetBar(args.Bars.Count - 1).Time)
                {
                    var footprintBar = new FootprintBar(this);
                    footprintBar.PopulatePrice(high, low, bid, ask, volume);
                    
                    base.AddBar(footprintBar, args, CalculateNewBarTime(time, args), open, high, low, close, volume, oi);
                }
                else
                {
                    if (args.Bars.GetBar(args.Bars.Count - 1) is FootprintBar footprintBar)
                    {
                        footprintBar.PopulatePrice(high, low, bid, ask, volume);     
                        
                        UpdateBar(args, footprintBar.Time, high, low, close, volume, oi);
                    }
                }
            }
            
            //and populate the global profile bars
            if (this.HasCumulativeProfile)
            {
                this.ProfileBar.PopulatePrice(high, low, bid, ask, volume);
            }
        }

        private DateTime CalculateNewBarTime(DateTime time, BarUpdateEventArgs args)
        {
            TimeSpan ts = TimeSpan.FromMinutes(this.BarValue);
            DateTime tmpTime = new DateTime(((time.Ticks + ts.Ticks - 1) / ts.Ticks) * ts.Ticks);

            if (tmpTime > args.SessionEnd)
                return args.SessionEnd;

            if (tmpTime < args.SessionStart)
                return args.SessionStart;

            return tmpTime;
        }
        
        protected override string CalculateBarSpan(Bars bars, DateTime dateTime)
        {
            if (bars.Count == 0)
                return "N/A";

            IBar bar = bars.GetBar(bars.Count - 1);
            if (bar.Time < dateTime)
                return TimeSpan.Zero.ToString("hh\\:mm\\:ss");

            return bar.Time.Subtract(dateTime).ToString("hh\\:mm\\:ss");
        }
        
        
        protected override void OnStateChange()
        {
            switch (this.State)
            {
                case SharpScript.State.Initialize:
                    break;
                case SharpScript.State.Configure:
                
                    break;
                case SharpScript.State.StartUp:
                    break;
                case SharpScript.State.Historical:
                    break;
                case SharpScript.State.Realtime:
                    break;
                case SharpScript.State.Finalize:
                    break;
                default:
                    break;
            }
        }

        public override Type[] ChartStyles => new Type[]
        {
            typeof(FootprintChartStyle) 
        };

        public override BuiltDataType[] SupportedBuiltDataTypes => new BuiltDataType[]
        {
            BuiltDataType.Minute, BuiltDataType.Tick
        };

        private BuiltDataType builtDataType = IO.BuiltDataType.Minute;
        public override BuiltDataType BuiltDataType { get => builtDataType; set => builtDataType = value; }


        private double barValue = 30;
        [Input]
        public double BarValue
        {
            get { return barValue; }
            set
            {
                barValue = Math.Max(1, value);
                NotifyPropertyChanged();
            }
        }
        
        private double tickSize = .05d;
        [Input]
        public double TickSize
        {
            get => tickSize;
            set => tickSize = Math.Max(double.Epsilon, value);
        }
        
        
        private bool hasCumulativeProfile = true;
        
        public bool HasCumulativeProfile
        {
            get => hasCumulativeProfile;
            set => hasCumulativeProfile = value;
        }
        
        
        [XmlIgnore]
        public double POCVolume { get; set; }
        
        private readonly FootprintBar profileBar;
        [XmlIgnore]
        public FootprintBar ProfileBar => this.profileBar;

        public override object Template => "FootprintBarType";
    }
}
