using ArthaChitra.Model.Common;
using ArthaChitra.Model.SharpScript;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using System.Xml.Serialization;

namespace ArthaChitra.Model.Chart
{
    [DisplayName("Footprint")]
    public sealed class FootprintChartStyle : ChartStyle
    {

        protected override void OnRender(DrawingContext dc, ChartVisualArgs args)
        {
            if (args.Bars.BarType is not FootprintBarType footprintBarType)
                return;
                
            if (args.Bars.Count == 0)
            {
                return;
            }
            
            double txtHeight = 10.0d; 

            for (int i = args.LastBarPainted; i >= args.FirstBarPainted; i--)
            {
                if (args.Bars.GetBar(i) is not FootprintBar bar)
                    continue;

                double highY = base.CalculateYFromPrice(bar.High, args);
                double lowY = base.CalculateYFromPrice(bar.Low, args);
                double closeY = base.CalculateYFromPrice(bar.Close, args);
                double openY = base.CalculateYFromPrice(bar.Open, args);

                Pen wickPen = this.GetOverridePen(args.Bars, i);
                Brush barBrush = this.GetOverrideBrush(args.Bars, i);
                double barHeight = 0.0d;
                double x = base.CalculateXFromBarIndex(i, args);
                double y = 0.0d;


                if (bar.Close >= bar.Open)
                {
                    barHeight = Math.Max(0.0d, openY - closeY);
                    y = closeY;

                    dc.DrawLine(wickPen, new Point(x, highY), new Point(x, closeY));
                    dc.DrawLine(wickPen, new Point(x, lowY), new Point(x, openY));
                }
                else
                {
                    barHeight = Math.Max(0.0d,closeY - openY);
                    y = openY;

                    dc.DrawLine(wickPen, new Point(x, highY), new Point(x, openY));
                    dc.DrawLine(wickPen, new Point(x, lowY), new Point(x, closeY));
                }

                dc.DrawRectangle(this.IsFilled? barBrush : null, wickPen, new Rect(x - (BarWidth / 2), y, BarWidth, barHeight));

                if (args.Bars.IsSelected && i % this.SelectionSteps(args) == 0)
                {
                    var midY = this.CalculateYFromPrice((bar.High + bar.Low) / 2.0d, args);
                    dc.DrawEllipse(args.ViewModel.Property.GridlinePen?.Brush ?? Brushes.Orange, args.ViewModel.Property.SessionPen, new Point(x, midY), Math.Max(5.0d, this.BarWidth), Math.Max(5.0d, this.BarWidth));
                }
                
                
                //draw the footprint objects....
                var poc = bar.POC;
                if (poc == null)
                    return;
                
                double max = Math.Min(args.Max, bar.Max);
                if (max == double.MinValue)
                    return;
                    
                double min = Math.Max(args.Min, bar.Min);
                if (min == double.MaxValue)
                    return;
                
                foreach (var kvp in bar.Items)
                {
                    if (kvp.Key < args.Min || kvp.Key > args.Max)
                        continue;
                        
                    y = CalculateYFromPrice(kvp.Key, args);
                    
                    var glyph = Globals.GetGlyphRun($"{kvp.Value.TotalVolume:N0}", new Point(x, y), AlignmentX.Left, AlignmentY.Center, Model.Common.Options.Config.FontSize, 96.0f, out var txtWidth, out txtHeight);                    
                    if (glyph != null)
                    {
                        if (kvp.Value.IsPointOfControl)
                        {
                            dc.DrawRectangle(null, Globals.DefaultForegroundPen, new Rect(x + this.BarWidth / 2.0d, y - txtHeight / 2.0d, txtWidth, txtHeight));
                        }
                        
                        //double width = Math.Max(1.0d, this.BarWidth * 5);   //width == 5 times of bar width
                        
                        var bidWidth = kvp.Value.TotalBidVolume / footprintBarType.POCVolume * this.ProfileWidth;
                        dc.DrawRectangle(this.BidBrush, null, new Rect(x - this.BarWidth / 2.0d - bidWidth, y - txtHeight / 3.0d, bidWidth, txtHeight / 3.0d * 2.0d));
                        var askWidth = kvp.Value.TotalAskVolume / footprintBarType.POCVolume * this.ProfileWidth;
                        dc.DrawRectangle(this.AskBrush, null, new Rect(x - this.BarWidth / 2.0d - bidWidth - askWidth, y - txtHeight / 3.0d, askWidth, txtHeight / 3.0d * 2.0d));
                        
                        var totalWidth = kvp.Value.TotalVolume / footprintBarType.POCVolume * this.ProfileWidth;
                        dc.DrawRectangle(null, Globals.DefaultForegroundPen, new Rect(x - this.BarWidth / 2.0d - totalWidth, y - txtHeight / 3.0d, totalWidth, txtHeight / 3.0d * 2.0d));
                        
                        dc.DrawText($"{kvp.Value.TotalVolume:N0}", new Point(x + this.BarWidth / 2.0d, y));
                    }
                }
                               
            }
            
            if (!footprintBarType.HasCumulativeProfile)
                return;
            
            
                
            var profileBar = footprintBarType.ProfileBar;
            if (profileBar != null)
            {
                var poc = profileBar.POC;
                if (poc != null && poc.TotalVolume > 0.0d)
                {
                    double price = Model.Data.Instrument.RoundToTickSize(args.Max, footprintBarType.TickSize);
                    while (price >= args.Min)
                    {
                        if (profileBar.Items.TryGetValue(price, out var item))
                        {
                            double y = this.CalculateYFromPrice(price, args);
                            
                            double width = Math.Max(this.ProfileWidth, args.ViewModel.Property.LeftMargin);
                            
                            var bidWidth = item.TotalBidVolume / poc.TotalVolume * width;
                            dc.DrawRectangle(this.BidBrush, null, new Rect(args.ViewModel.LeftYAxisWidth, y - txtHeight / 3.0d, bidWidth, txtHeight / 3.0d * 2.0d));
                            var askWidth = item.TotalAskVolume / poc.TotalVolume * width;
                            dc.DrawRectangle(this.AskBrush, null, new Rect(args.ViewModel.LeftYAxisWidth + bidWidth, y - txtHeight / 3.0d, askWidth, txtHeight / 3.0d * 2.0d));
                            
                            dc.DrawRectangle(null, Globals.DefaultForegroundPen, new Rect(args.ViewModel.LeftYAxisWidth, y - txtHeight / 3.0d, item.TotalVolume / poc.TotalVolume * width, txtHeight / 3.0d * 2.0d)); 
                        }
                    
                        price = Model.Data.Instrument.RoundToTickSize(price - footprintBarType.TickSize, footprintBarType.TickSize);
                    }
                }
            } 

        } 

        protected override void OnStateChange()
        {
            switch (this.State)
            {
                case SharpScript.State.Initialize:
                    this.BidBrush = Brushes.Aqua;
                    this.AskBrush = Brushes.HotPink;
                    break;
                case SharpScript.State.Configure:
                    break;
                case SharpScript.State.StartUp:
                    break;
                case SharpScript.State.Historical:
                    break;
                case SharpScript.State.Realtime:
                    break;
                case SharpScript.State.Finalize:
                    break;
                default:
                    break;
            }
        }

        
        private bool isFilled = true;

        public bool IsFilled
        {
            get { return isFilled; }
            set
            {
                isFilled = value;
                NotifyPropertyChanged();
            }
        }
        
        
        private double profileWidth = 30.0d;
        
        public double ProfileWidth
        {
            get => profileWidth;
            set => profileWidth = Math.Max(0, value);
        }
        
        
        private Brush bidBrush;
        [XmlIgnore]
        public Brush BidBrush
        {
            get => bidBrush;
            set
            {
                bidBrush = value;
                if (bidBrush != null && bidBrush.CanFreeze)
                    bidBrush.Freeze();
                    
                NotifyPropertyChanged();
            }
        }
        
        [XmlElement(ElementName = "BidBursh")]
        public SerializerHelper<Brush> BidBrushSerializer
        {
            get => new SerializerHelper<Brush>(this.BidBrush);
            set => this.BidBrush = value?.Data;
        }


        private Brush askBrush;
        [XmlIgnore]
        public Brush AskBrush
        {
            get => this.askBrush;
            set
            {
                askBrush = value;
                if (askBrush != null && askBrush.CanFreeze)
                    askBrush.Freeze();
                    
                NotifyPropertyChanged();
            }
        }
        
        [XmlElement(ElementName = "AskBrush")]
        public SerializerHelper<Brush> AskBrushSerializer
        {
            get => new SerializerHelper<Brush>(this.AskBrush);
            set => this.AskBrush = value?.Data;
        }
        

        public override object Template => this.GetType().Name;

    }
}
